# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Get charges using GULP

"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import numpy as np
import subprocess as sp
from soprano.utils import safe_communicate
from soprano.calculate.gulp._utils import (_gulp_cell_definition,
                                           _gulp_parse_charges)


def get_gulp_charges(s, charge_method="eem", save_charges=True,
                     gulp_command='gulp',
                     gulp_path=None):
    """Calculate the atomic partial charges using GULP.

    | Parameters:
    |   s (ase.Atoms): the structure to calculate the energy of
    |   charge_method (Optional[str]): which method to use for atomic partial
    |                                  charge calculation. Can be any of
    |                                  'eem', 'qeq' and 'pacha'.
    |                                  Default is 'eem'.
    |   save_charges (Optional[bool]): whether to save or not the charges in
    |                                  the given ase.Atoms object. Default is
    |                                  True.
    |   gulp_command (Optional[str]): command required to call the GULP
    |                                 executable.
    |   gulp_path (Optional[str]): path where the GULP executable can be
    |                              found. If not present, the GULP command
    |                              will be invoked directly (assuming the
    |                              executable is in the system PATH).

    | Returns:
    |   charges(np.array(float)): per-atom partial charges

    """

    # Sanity check
    if charge_method not in ['eem', 'qeq', 'pacha']:
        raise ValueError('Invalid charge_method passed to get_gulp_charges')

    # Now define the input
    gin = "{0}\n".format(charge_method)
    gin += _gulp_cell_definition(s)

    # AND GO!
    if gulp_path is None:
        gulp_path = ''

    gulp_cmd = [os.path.join(gulp_path, gulp_command)]

    # Run the thing...
    try:
        gulp_proc = sp.Popen(gulp_cmd,
                             universal_newlines=True,
                             stdin=sp.PIPE,
                             stdout=sp.PIPE,
                             stderr=sp.PIPE)
        stdout, stderr = safe_communicate(gulp_proc, gin)
    except OSError:
        raise RuntimeError('GULP not found on this system with the given '
                           'command')

    # Necessary for compatibility in Python2
    try:
        stdout = unicode(stdout)
    except NameError:
        pass

    # And parse the output
    gulp_lines = stdout.split('\n')
    charges = _gulp_parse_charges(gulp_lines)
    if charges is None:
        raise RuntimeError('ERROR - GULP run failed to return charges')

    # Run a security check
    if not np.all(s.get_atomic_numbers() == charges['Z']):
        raise RuntimeError('ERROR - Invalid charges parsed from GULP output')

    charges = charges['q']

    if save_charges:
        s.set_initial_charges(charges)

    return charges



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes and functions for using the W99 force field in GULP.
This force field only applies to organic molecules. More information can be
found in the original paper by Donald E. Williams:

D.E. Williams,
*Improved Intermolecular Force Field for Molecules Containing H, C, N, and O
Atoms, with Application to Nucleoside and Peptide Crystals*  - Journal
of Computational Chemistry, Vol. 22, No. 11, 1154-1166 (2001)

"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re
import os
import json
import pkgutil
import numpy as np
import subprocess as sp
from ase.calculators.singlepoint import SinglePointCalculator
from soprano.utils import safe_communicate
from soprano.properties.linkage import Molecules
from soprano.calculate.gulp._utils import (_gulp_cell_definition,
                                           _gulp_parse_energy,
                                           _gulp_parse_charges)

_w99_data = pkgutil.get_data('soprano',
                             'data/w99_parameters.json').decode('utf-8')
_w99_data = json.loads(_w99_data)


class W99Error(Exception):
    pass


def find_w99_atomtypes(s, force_recalc=False):
    """Calculate the W99 force field atom types for a given structure.

    | Parameters:
    |   s (ase.Atoms): the structure to calculate the atomtypes on
    |   force_recalc (bool): whether to recalculate the molecules even if
    |                        already present. Default is False.
    """

    # First, check that W99 even applies to this system
    chsyms = np.array(s.get_chemical_symbols())

    if not set(chsyms).issubset(set(['C', 'H', 'O', 'N'])):
        raise W99Error('Invalid system for use of W99 force field'
                       ' - System can only contain H, C, O and N!')

    if Molecules.default_name not in s.info or force_recalc:
        Molecules.get(s)

    mols = s.info['molecules']

    # Now the types!
    w99types = np.array(['XXX' for sym in chsyms])

    for m_i, m in enumerate(mols):

        inds = list(m.indices)
        bonds = m.get_array('bonds')
        msyms = chsyms[inds]

        for a_mi, a_i in enumerate(inds):
            # What element is it?
            el_i = msyms[a_mi]
            bnd_i = bonds[a_mi]
            if len(bnd_i) < 1:
                # Something's wrong
                raise W99Error('ERROR - W99 can not be applied to single'
                               ' atom molecular fragments')
            if el_i == 'H':
                # Hydrogen case
                if len(bnd_i) != 1:
                    raise W99Error('ERROR - Hydrogen with non-1 valence '
                                   'found')
                # What is it bonded to?
                nn_el = chsyms[bnd_i[0]]
                if nn_el == 'C':
                    w99types[a_i] = 'H_1'
                elif nn_el == 'N':
                    w99types[a_i] = 'H_4'
                elif nn_el == 'O':
                    # Alcoholic or carboxylic?
                    nn_mi = inds.index(bnd_i[0])
                    carbs = [b for b in bonds[nn_mi] if chsyms[b] == 'C']
                    if len(carbs) != 1:
                        raise W99Error('ERROR - Anomalous chemical group '
                                       'found')
                    c_mi = inds.index(carbs[0])
                    oxys = [b for b in bonds[c_mi]
                            if chsyms[b] == 'O' and b != bnd_i[0]]
                    # What is it?
                    if len(oxys) == 1 and len(bonds[c_mi]) == 3:
                        # Carboxylic!
                        w99types[a_i] = 'H_3'
                    else:
                        # Alcoholic!
                        w99types[a_i] = 'H_2'
            elif el_i == 'C':
                # Carbon case
                val = len(bnd_i)
                if val > 1 and val < 5:
                    w99types[a_i] = 'C_{0}'.format(val)
                else:
                    raise W99Error('ERROR - Anomalous chemical group found')
            elif el_i == 'O':
                # Oxygen case
                val = len(bnd_i)
                if val > 0 and val < 3:
                    w99types[a_i] = 'O_{0}'.format(val)
                else:
                    raise W99Error('ERROR - Anomalous chemical group found')
            elif el_i == 'N':
                # Nitrogen case
                val = len(bnd_i)
                if val == 3:
                    w99types[a_i] = 'N_1'
                else:
                    # Count hydrogens
                    hydros = [b for b in bnd_i if chsyms[b] == 'H']
                    if len(hydros) == 0:
                        w99types[a_i] = 'N_2'
                    elif len(hydros) == 1:
                        w99types[a_i] = 'N_3'
                    else:
                        w99types[a_i] = 'N_4'

        m.set_array('w99_types', w99types[inds])

    s.set_array('w99_types', w99types)


def _w99_field_definition(s, etol):
    """Output the W99 field definition for GULP input. System must already
    have w99 types calculated"""

    w99types = s.get_array('w99_types')

    # Which types are present?
    w99types = list(set(w99types))

    field_def = ''
    # Now take care of all permutations
    for i, t1 in enumerate(w99types):
        for t2 in w99types[i:]:
            # Calculate parameters
            if t1 == t2:
                A, rho, C = _w99_data[t1]
            else:
                A1, rho1, C1 = _w99_data[t1]
                A2, rho2, C2 = _w99_data[t2]
                A = np.sqrt(A1*A2)
                rho = 2.0/(1.0/rho1+1.0/rho2)
                C = np.sqrt(C1*C2)

            # Calculate the optimal cutoff to meet the required energy
            # tolerance
            expcut = -rho*np.log(etol/abs(A))
            r6cut = (abs(C)/etol)**(1.0/6.0)
            cut = max(expcut, r6cut, 0)
            field_def += 'buck inter\n{0} {1} {2} {3} {4} {5}\n'.format(t1,
                                                                        t2,
                                                                        A,
                                                                        rho,
                                                                        C,
                                                                        cut)

    return field_def


def get_w99_energy(s, charge_method='eem', Etol=1e-6,
                   gulp_command='gulp',
                   gulp_path=None,
                   save_charges=False):
    """Calculate the W99 force field energy using GULP.

    | Parameters:
    |   s (ase.Atoms): the structure to calculate the energy of
    |   charge_method (Optional[str]): which method to use for atomic partial
    |                                  charge calculation. Can be any of
    |                                  'eem', 'qeq' and 'pacha'.
    |                                  Default is 'eem'.
    |   Etol (Optional[float]): tolerance on energy for intermolecular
    |                           potential cutoffs (relative to single
    |                           interaction energy). Default is 1e-6 eV.
    |   gulp_command (Optional[str]): command required to call the GULP
    |                                 executable.
    |   gulp_path (Optional[str]): path where the GULP executable can be
    |                              found. If not present, the GULP command
    |                              will be invoked directly (assuming the
    |                              executable is in the system PATH).
    |   save_charges (Optional[bool]): whether to retrieve also the charges
    |                                  and save them in the Atoms object.
    |                                  False by default.

    | Returns:
    |   energy (float): the calculated energy

    """

    # Sanity check
    if charge_method not in ['eem', 'qeq', 'pacha']:
        raise ValueError('Invalid charge_method passed to get_w99_energy')

    # First, atom types
    find_w99_atomtypes(s)

    # Now define the input
    gin = "molq {0} dipole\n".format(charge_method)
    gin += _gulp_cell_definition(s, syms=s.get_array('w99_types'))

    # Finally, the potential definition
    gin += _w99_field_definition(s, Etol)

    # AND GO!
    if gulp_path is None:
        gulp_path = ''

    gulp_cmd = [os.path.join(gulp_path, gulp_command)]

    try:
        gulp_proc = sp.Popen(gulp_cmd,
                             universal_newlines=True,
                             stdin=sp.PIPE,
                             stdout=sp.PIPE,
                             stderr=sp.PIPE)
        stdout, stderr = safe_communicate(gulp_proc, gin)
    except OSError:
        raise RuntimeError('GULP not found on this system with the given '
                           'command')

    # Necessary for compatibility in Python2
    try:
        stdout = unicode(stdout)
    except NameError:
        pass

    # Now parse the energy
    gulp_lines = stdout.split('\n')
    E = _gulp_parse_energy(gulp_lines)
    if E is None:
        raise RuntimeError('ERROR - GULP run failed to return energy')

    # Remember it with a mock ASE calculator
    calc = SinglePointCalculator(s, energy=E)
    s.set_calculator(calc)

    if save_charges:
        qs = _gulp_parse_charges(gulp_lines)
        s.set_initial_charges(qs['q'])

    return E



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Utility functions for symmetry calculations"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


import numpy as np
from scipy.linalg import null_space
from soprano.optional import requireSpglib


@requireSpglib('spglib')
def _get_symmetry_dataset(s, symprec, spglib=None):
    lattice = s.get_cell()
    positions = s.get_scaled_positions()
    numbers = s.get_atomic_numbers()
    symdata = spglib.get_symmetry_dataset((lattice, positions, numbers),
                                          symprec=symprec)
    return symdata


@requireSpglib('spglib')
def _get_symmetry_ops(hall_no, spglib=None):
    symdata = spglib.get_symmetry_from_database(hall_no)
    return symdata['rotations'], symdata['translations']


def _loci_intersect(l1, l2):
    """Find the locus of points at the intersection of l1 and l2"""

    # Make sure l1 has the lowest dof
    l1, l2 = sorted([l1, l2], key=lambda x: x[0])

    if l1[0] == 0 or l2[0] == 3:
        return l1
    elif l1[0] == 1:
        d = np.dot(l1[1], l2[1])
        if l2[0] == 1:
            if abs(abs(d) - 1) < 1e-5:
                return (1, l1[1])
            else:
                return (0, None)
        else:
            if abs(d) > 1e-5:
                return (0, None)
            else:
                return (1, l1[1])
    elif l1[0] == 2:
        # l2 has to be a plane too
        v = np.cross(l1[1], l2[1])
        return (1, v/np.linalg.norm(v))


def _find_wyckoff_points(a, symprec=1e-5):
    """Find and return all Wyckoff points for a given atomic system,
    as well as the operations that each Wyckoff point is stable
    under, and whether the Hessian has local radial symmetry or is
    definite in them."""

    dset = _get_symmetry_dataset(a, symprec)
    hno = dset['hall_number']

    # First, check the standard cell
    stdcell = dset['std_lattice']
    # And the transformation properties
    icell = np.linalg.inv(stdcell)
    ic2 = np.dot(icell.T, icell)
    evals, _ = np.linalg.eig(ic2)
    e0 = evals[0]
    is_iso = np.isclose(evals, e0).all()
    is_def = (np.sign(evals) == np.sign(e0)).all()

    # Operations in the non-transformed frame
    R, T = _get_symmetry_ops(hno)
    # Transformation
    P, o = dset['transformation_matrix'], dset['origin_shift']
    iP = np.linalg.inv(P)

    # Operations in the transformed frame
    rR = np.einsum('ij,ajk,kl', iP, R, P)
    rT = np.dot(T, iP.T) % 1

    # Invariant loci for the operations
    invLoci = []
    for ro in R:
        evals, evecs = np.linalg.eig(ro)
        ev1 = np.isclose(evals, 1)
        dof = np.sum(ev1)
        if dof == 0:
            v = None
        elif dof == 1:
            i = np.where(ev1)[0][0]
            v = np.real(evecs[:, i])
        elif dof == 2:
            i = np.where(ev1)[0]
            v = np.real(np.cross(*evecs[:, i].T))
            v /= np.linalg.norm(v)
        elif dof == 3:
            v = None
        else:
            raise RuntimeError('Invalid symmetry operation')
        invLoci.append((dof, v))

    invLoci = np.array(invLoci, dtype=object)

    # A grid of candidate Wyckoff points: they always come in
    # fractions n/24
    wgrid = np.reshape(np.meshgrid(*[range(24)]*3, indexing='ij'),
                       (3, -1)).T

    # Identify the points that match under transformation and modulo 24
    matching_ops = np.all(np.isclose((np.tensordot(R, wgrid,
                                                   axes=(2, 1)) +
                                      T[:, :, None]*24) % 24,
                                     wgrid.T[None] % 24),
                          axis=1)
    # Indices of all matching points
    w_indices = np.where(np.any(matching_ops[1:], axis=0))[0]

    # Now for each find how many degrees of freedom there are
    wp_ops = []
    wp0_ops = []
    wp_indices = []
    for i in w_indices:
        ops_inds = np.where(matching_ops[:, i])[0]
        loci = sorted(invLoci[ops_inds], key=lambda x: x[0])
        ltot = loci[0]
        for l in loci[1:]:
            if ltot[0] == 0:
                break
            ltot = _loci_intersect(ltot, l)
        if ltot[0] == 0:
            wp_indices.append(i)
            wp0_ops.append(zip(R[ops_inds], T[ops_inds]))
            wp_ops.append(zip(rR[ops_inds], rT[ops_inds]))

    # Find their positions in fractional coordinates
    wp_fxyz = (np.dot(wgrid[wp_indices]/24.0, iP.T)-np.dot(iP, o)) % 1
    # Remove any identical rows
    if wp_fxyz.shape[0] > 0:
        wp_fxyz, uinds = np.unique(np.round(wp_fxyz*24).astype(int), axis=0, return_index=True)
        wp_fxyz = wp_fxyz/24.0
    else:
        uinds = []
    wp0_ops = np.array(wp0_ops)[uinds]
    wp_ops = np.array(wp_ops)[uinds]

    isohess = [] # Here the meanings are:
    # 0 - no constraints
    # 1 - is positive/negative definite
    # 2 - is isotropic
    for ops in wp0_ops:
        wh = _wyckoff_isohess(ops)
        isohess.append(wh*(2*is_def+is_iso-1))
    isohess = np.array(isohess)

    return wp_fxyz, wp_ops, isohess


def _wyckoff_isohess(ops):
    """For each set of operations in wp_ops, representing a Wyckoff
    point that is symmetric under them, find whether the symmetry is
    such that any symmetric 3x3 tensor valued function ought to be 
    isotropic in that point"""

    """
    How it works:
     - For any operation, tensors must respect the following:

                    T = O'TO

        with O' meaning the transpose. This can be rewritten as

                    kron(O', O')t = t

       where t is a reshaped vector version of T. In other words,
       t has to belong to the nullspace of kron(O', O')-I.
     - While kron(O', O') is a 9x9 matrix, if we assume that T is
       symmetric, then it has 6 independent components. We want to
       prove that T is isotropic (T = xI, with x a scalar), which
       means we only worry about 5 independent components.
     - We can split T as such:

                    T = xI + S = xO'O + O'SO

       where S is a traceless symmetric tensor. Now, in what 
       conditions can we say that S has to be zero?
     - The first is that O'O = I. This is true of rotations
       and reflections, but not of all symmetry operations.
       If it's not true for any operation in wp_ops, then S
       can not be zero.
     - The second is that the intersection of the null
       spaces of P2kron(O',O')P1 is empty. Here P1 is a 9x5 
       matrix such that P1s = t-xi, where s is the vectorised
       form of S and i the same for I, and P2 is the inverse.

    Note that this has to hold for the operations expressed in an *orthogonal*
    basis. Otherwise it can fail. However, if it works in one basis, it's 
    valid in all of them. 
    """

    # Rotation/transformation matrices
    R = list(zip(*ops))[0]

    # First: check that all matrices are orthogonal
    ortho = all([np.isclose(np.dot(r.T, r), np.eye(3)).all() for r in R])

    if not ortho:
        return False

    # These matrices go back and fro to the 'reduced' form of the
    # unfolded symmetric traceless tensor (9 to only 5 components)

    P1 = np.array([[1, 0, 0, 0, 0],
                   [0, 1, 0, 0, 0],
                   [0, 0, 1, 0, 0],
                   [0, 1, 0, 0, 0],
                   [0, 0, 0, 1, 0],
                   [0, 0, 0, 0, 1],
                   [0, 0, 1, 0, 0],
                   [0, 0, 0, 0, 1],
                   [-1, 0, 0, -1, 0]])

    P2 = np.array([
        [1, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 1, 0, 0, 0],
    ])

    ns = np.eye(5)
    for r in R:
        r5 = np.linalg.multi_dot([P2, np.kron(r, r).T, P1])
        ns1 = null_space(r5-np.eye(5))
        ns = np.dot(ns, null_space(np.concatenate([ns, -ns1],
                                                  axis=1))[:ns.shape[1]])
        if ns.shape[1] == 0:
            return True
        ns /= np.linalg.norm(ns, axis=0)[None, :]

    return False



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
selection.py

Contains the definition of an AtomSelection class,
namely a group of selected atoms for a given structure,
and methods to build it.
"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import copy
import hashlib
import warnings
import operator
import numpy as np

from soprano.utils import minimum_supcell, supcell_gridgen, customize_warnings

customize_warnings()

# This decorator applies to all operators providing some basic checksP


def _operator_checks(opfunc):

    def decorated_opfunc(self, other):
        if not isinstance(other, AtomSelection):
            raise TypeError('AtomSelection does not support operations with'
                            ' different types')

        if self._auth is not None and other._auth is not None:
            # Check compatibility
            if self._auth != other._auth:
                raise ValueError('Selections come from different systems')

        return opfunc(self, other)

    return decorated_opfunc


class AtomSelection(object):

    """AtomSelection object.

    An AtomSelection represents a group of atoms from an ASE Atoms object.
    It keeps track of them and can be used to perform operations on them
    (for example geometrical transformation or extraction of specific
    properties).
    It does not keep track of the original Atoms object it's been created
    from, but can be "authenticated" to verify that it is indeed operating
    consistently on the same structure. It also provides a series of static
    methods to build selections with various criteria.

    """

    def __init__(self, atoms, sel_indices, authenticate=True):
        """Initialize the AtomSelection.

        | Args:
        |   atoms (ase.Atoms): the atoms object on which the selection is
        |                      applied
        |   sel_indices (list[int]): the list of indices of the atoms that
        |                            are to be selected
        |   authenticate (Optional[bool]): whether to use hashing to confirm
        |                                  the identity of the atoms object
        |                                  we're operating with

        """

        # A quick check: are the indices actually contained in the Atoms?
        if len(sel_indices) > 0:
            if (min(sel_indices) < 0 or
                    max(sel_indices) >= len(atoms)):
                raise ValueError('Invalid indices for given Atoms object')

        self._indices = np.array(sel_indices)

        if authenticate:
            # Create an hash for certification
            self._auth = self._hash(atoms)
        else:
            self._auth = None

        self._arrays = {}

    @property
    def indices(self):
        return self._indices

    def _hash(self, atoms):
        """A function to create an identifying hash for a given Atoms system.
        This is used later to check that the system is indeed unchanged when
        the Selection is reused.

        While changes in positions or cell don't invalidate the Selection,
        changes in composition potentially do (indices of atoms can change).
        """

        h = hashlib.md5()
        h.update(''.join(atoms.get_chemical_symbols()).encode())

        return h.hexdigest()

    def has(self, name):
        """Check if the selection has a given array

        | Args:
        |   name (str): name of the array to be checked for

        | Returns:
        |   has (bool): if the array is present or not
        """

        return name in self._arrays

    def set_array(self, name, array):
        """Save an array of given name containing arbitraty information
        tied to the selected atoms.
        This must match the length of the selection and will be passed on to
        any Atoms objects created with .subset.

        | Args:
        |   name (str): name of the array to be set or created
        |   array (np.ndarray): array of data to be saved

        """

        # First a check
        if len(array) != len(self):
            raise ValueError("Invalid array passed to set_array")

        self._arrays[name] = np.array(array)

    def get_array(self, name):
        """Retrieve a previously stored data array.

        | Args:
        |   name (str): name of the array to be set or created

        | Returns:
        |   array (np.ndarray): array of data to be saved

        """

        # If the name isn't right just let the KeyError happen
        return self._arrays[name]

    def validate(self, atoms):
        """Check that the given Atoms object validates with this selection."""
        if self._auth is None:
            warnings.warn('WARNING'
                          ' - this selection does not support validation')
            return True
        else:
            return self._hash(atoms) == self._auth

    def subset(self, atoms, use_cell_indices=False):
        """Generate an Atoms object containing only the selected atoms.

        | Args:
        |   atoms (ase.Atoms):       Atoms object from which to take the
        |                            selection
        |   use_cell_indices (bool): If True, use the cell_indices array to
        |                            pick the specified periodic copies of
        |                            the corresponding atoms (useful e.g. to
        |                            take the correct periodic copies for a
        |                            molecule)

        | Returns:
        |   subset (ase.Atoms):      Atoms object containing only the
        |                            specified selection
        """

        if not self.validate(atoms):
            raise ValueError(
                'Given Atoms object does not match this selection')

        subset = atoms[self._indices]
        # Copy any extra arrays
        for k, arr in self._arrays.items():
            subset.set_array(k, arr.copy())

        if use_cell_indices and subset.has('cell_indices'):
            ijk = subset.get_array('cell_indices')
            subset.set_scaled_positions(subset.get_scaled_positions() + ijk)

        return subset

    def __getitem__(self, indices):
        """Slicing: take only part of a selection"""

        if type(indices) is int:
            # Special case, a single element!
            indices = slice(indices, indices+1)

        try:
            newsel = self._indices[indices]
        except TypeError:
            newsel = [self._indices[i] for i in indices]

        sliced = copy.deepcopy(self)
        sliced._indices = newsel
        sliced._arrays = {k: a[indices] for k, a in self._arrays.items()}

        return sliced

    def __iter__(self):
        for i in range(len(self._indices)):
            yield self[i]

    # Overloading operators to allow sum, subtraction and product of selections
    @_operator_checks
    def __add__(self, other):
        """Sum: join selections"""

        # Join
        ans = copy.deepcopy(self)
        ans._indices = np.array(list(set(self.indices).union(other.indices)))
        # For the arrays:
        # only join the ones present in BOTH selections
        common_k = set(self._arrays.keys()
                       ).intersection(set(other._arrays.keys()))
        ans._arrays = {}
        for k in common_k:
            ans._arrays[k] = np.concatenate((self._arrays[k],
                                             other._arrays[k]))

        return ans

    @_operator_checks
    def __sub__(self, other):

        # Difference
        ans = copy.deepcopy(self)
        ans._indices = np.array(list(set(self.indices)-set(other.indices)))
        # For the arrays:
        # keep them but remove the removed indices
        arr_i = [np.where(self.indices == i)[0][0] for i in ans._indices]
        for k in ans._arrays:
            ans._arrays[k] = ans._arrays[k][arr_i]

        return ans

    @_operator_checks
    def __mul__(self, other):

        # Intersection
        ans = copy.deepcopy(self)
        ans._indices = np.array(list(set(self.indices)
                                     .intersection(other.indices)))
        # For the arrays:
        # keep the ones present in either selection,
        # but only the relevant indices of course,
        # and remove if conflicting!
        all_k = set(self._arrays.keys()).union(set(other._arrays.keys()))
        arr1_i = [np.where(self.indices == i)[0][0] for i in ans._indices]
        arr2_i = [np.where(other.indices == i)[0][0] for i in ans._indices]
        ans._arrays = {}
        for k in all_k:
            try:
                arr1 = self._arrays[k][arr1_i]
            except KeyError:
                arr1 = None
            try:
                arr2 = other._arrays[k][arr2_i]
            except KeyError:
                arr2 = None

            if arr1 is not None and arr2 is not None:
                # Do they conflict?
                if not np.all(arr1 == arr2):
                    print(('WARNING - conflicting arrays of name {0} found'
                           ' will be removed during intersection'
                           ' operation').format(k))
                    continue

            ans._arrays[k] = arr1 if arr1 is not None else arr2

        return ans

    def __len__(self):
        return len(self._indices)

    def __contains__(self, item):
        return item in self._indices

    @staticmethod
    def all(atoms):
        """Generate a selection for the given Atoms object of all atoms.

        | Args:
        |   atoms (ase.Atoms): Atoms object on which to perform selection

        | Returns:
        |   selection (AtomSelection)

        """

        return AtomSelection(atoms, range(len(atoms)))

    @staticmethod
    def from_element(atoms, element):
        """Generate a selection for the given Atoms object of all atoms of a
        specific element.

        | Args:
        |   atoms (ase.Atoms): Atoms object on which to perform selection
        |   element (str): symbol of the element to select

        | Returns:
        |   selection (AtomSelection)

        """

        sel_i = np.where(np.array(atoms.get_chemical_symbols()) == element)[0]

        return AtomSelection(atoms, sel_i)

    @staticmethod
    def from_box(atoms, abc0, abc1, periodic=False, scaled=False):
        """Generate a selection for the given Atoms object of all atoms within
        a given box volume.

        | Args:
        |   atoms (ase.Atoms): Atoms object on which to perform selection
        |   abc0 ([float, float, float]): bottom corner of box
        |   abc1 ([float, float, float]): top corner of box
        |   periodic (Optional[bool]): if True, include periodic copies of the
        |                              atoms
        |   scaled (Optional[bool]): if True, consider scaled (fractional)
        |                            coordinates instead of absolute ones

        | Returns:
        |   selection (AtomSelection)

        """

        if scaled:
            pos = atoms.get_scaled_positions()
        else:
            pos = atoms.get_positions()
        # Do we need periodic copies?
        if periodic and any(atoms.get_pbc()):
            # Get the range
            max_r = np.linalg.norm(np.array(abc1)-abc0)
            scell_shape = minimum_supcell(max_r, latt_cart=atoms.get_cell(),
                                          pbc=atoms.get_pbc())
            grid_frac, grid = supcell_gridgen(atoms.get_cell(), scell_shape)
            if scaled:
                pos = (pos[:, None, :]+grid_frac[None, :, :])
            else:
                pos = (pos[:, None, :]+grid[None, :, :])

        where_i = np.where(np.all(pos > abc0, axis=-1) &
                           np.all(pos < abc1, axis=-1))[:2]

        sel_i = where_i[0]

        sel = AtomSelection(atoms, sel_i)
        if periodic:
            sel.set_array('cell_indices', grid_frac[where_i[1]])

        return sel

    @staticmethod
    def from_sphere(atoms, center, r, periodic=False, scaled=False):
        """Generate a selection for the given Atoms object of all atoms within
        a given spherical volume.

        | Args:
        |   atoms (ase.Atoms): Atoms object on which to perform selection
        |   center ([float, float, float]): center of the sphere
        |   r (float): radius of the sphere
        |   periodic (Optional[bool]): if True, include periodic copies of the
        |                              atoms
        |   scaled (Optional[bool]): if True, consider scaled (fractional)
        |                            coordinates instead of absolute ones

        | Returns:
        |   selection (AtomSelection)

        """

        if scaled:
            pos = atoms.get_scaled_positions()
        else:
            pos = atoms.get_positions()
        # Do we need periodic copies?
        if periodic and any(atoms.get_pbc()):
            # Get the range
            r_bounds = minimum_supcell(r, latt_cart=atoms.get_cell(),
                                       pbc=atoms.get_pbc())
            grid_frac, grid = supcell_gridgen(atoms.get_cell(), r_bounds)
            if scaled:
                pos = (pos[:, None, :]+grid_frac[None, :, :])
            else:
                pos = (pos[:, None, :]+grid[None, :, :])

        where_i = np.where(np.linalg.norm(pos-center, axis=-1) <= r)

        sel_i = where_i[0]

        sel = AtomSelection(atoms, sel_i)
        if periodic:
            sel.set_array('cell_indices', grid_frac[where_i[1]])

        return sel

    @staticmethod
    def from_bonds(atoms, center, n, op='le'):
        """Generate a selection for the given Atoms object of other atoms
        based on their reciprocal bonding distance. Default is selection of
        all atoms that are within a certain bonding distance (less-or-equal
        than n). However different operators can be specified for different
        selection criteria. Atoms that do not belong to the same tree of the
        bonding graph are never selected.

        | Args:
        |   atoms (ase.Atoms): Atoms object on which to perform selection
        |   center (int): index of the atom to compute the bonding distance
        |                 from
        |   n (int): bonding distance to compare
        |   op (Optional[str]): operator to use for comparison with the given
        |                       bonding distance. By default it's le, meaning
        |                       "less or equal" than n, which means all atoms
        |                       will be selected that are at most n bonds away
        |                       from the center.
        |                       Other options are the functions present in the
        |                       `operator` module and are:
        |                             - lt : less than
        |                             - le : less or equal
        |                             - eq : exactly equal
        |                             - ge : greater or equal
        |                             - gt : greater than
        """

        # Start by computing the bonding graph
        from soprano.properties.linkage import BondGraph, Bonds
        from soprano.utils import get_bonding_distance

        bgraph = BondGraph.get(atoms)
        op = getattr(operator, op)
        sel_i = []

        for i in np.arange(len(atoms)):
            d = get_bonding_distance(bgraph, center, i)
            if d > -1 and op(d, n):
                sel_i.append(i)

        return AtomSelection(atoms, sel_i)

    @staticmethod
    def from_array(atoms, name, value, op='eq'):
        """Generate a selection for the given Atoms object of other atoms
        based on a comparison with some array value. Default is selection of
        all atoms that have the same exact value. However different operators
        can be specified for different selection criteria.

        | Args:
        |   atoms (ase.Atoms): Atoms object on which to perform selection
        |   name (str): name of the array to select with
        |   value (any type): value to compare the contents of the array with
        |   op (Optional[str]): operator to use for comparison with the given
        |                       value. By default it's eq, meaning
        |                       "equal" to value, which means all atoms
        |                       will be selected for whose the array of given
        |                       name has the given value.
        |                       Other options are the functions present in the
        |                       `operator` module and are:
        |                             - lt : less than
        |                             - le : less or equal
        |                             - eq : exactly equal
        |                             - ge : greater or equal
        |                             - gt : greater than
        """

        arr = atoms.get_array(name)
        op = getattr(operator, op)
        sel_i = np.where(op(arr, value))[0]

        return AtomSelection(atoms, sel_i)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implementation of AtomsProperties that relate to bond order parameters"""

# Python 2-to-3 compatibility code
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
from scipy.special import sph_harm
from soprano.utils import wigner_3j, minimum_periodic
from soprano.properties import AtomsProperty
from soprano.selection import AtomSelection


def _valid_triples(l):
    """Valid triples for angular momentum l for which m1 + m2 + m3 = 0"""

    # m1 can be anything
    m1 = np.arange(-l, l+1)
    # m2 can be anything such that |m1+m2| <= l
    m2 = [np.arange(-l-min(m, 0), l+1-max(m, 0)) for m in m1]

    return np.array([[m, mm, -m-mm] for i, m in enumerate(m1)
                     for mm in m2[i]])


def _steinhardt_pars(vecs, l_channels, compute_W=False, weights=None):
    """Compute the Steinhardt bond order parameters (Q and optionally W)"""

    Qlm = []
    Q = np.zeros(len(l_channels))
    W = np.zeros(len(l_channels))

    # Translate vectors in polar form
    vecs = np.array(vecs)/np.linalg.norm(vecs, axis=1)[:, None]
    pols = np.array([np.arccos(vecs[:, 2]),
                     np.arctan2(vecs[:, 1], vecs[:, 0])]).T

    # What about weights?
    if weights is None:
        weights = np.ones(vecs.shape[0])/len(vecs)

    for i, l in enumerate(l_channels):
        mvals = np.arange(-l, l+1)
        Qlm.append(np.sum(sph_harm(mvals[None, :], l, pols[:, 1, None],
                                   pols[:, 0, None])*weights[:, None], axis=0))
        Qsum = np.sum(np.conj(Qlm[i])*Qlm[i])
        Q[i] = np.abs(np.sqrt(Qsum*4*np.pi/(2*l+1.0)))
        if compute_W:
            m1, m2, m3 = _valid_triples(l).T
            ls = np.array([l]*len(m1))
            w3j = wigner_3j(ls, m1, ls, m2, ls, m3)
            W[i] = np.abs(np.sum(w3j*Qlm[i][m1+l]*Qlm[i][m2+l]*Qlm[i][m3+l]
                                 )/(Qsum)**1.5)

    if not compute_W:
        return Q
    else:
        return Q, W


class BondOrder(AtomsProperty):

    """
    BondOrder

    Computes the order parameters defined by Steinhardt et al. [PRB 28, 784
    (1983)] for an atom or group of atoms in the system. Here for each atom
    all closest periodic copies of the others are used; a sigmoidal cutoff is
    used to reduce the impact of those beyond a certain distance while keeping
    the behaviour smooth upon small fluctuations.

    | Parameters:
    |   l_channels (list[int]): list of angular momentum channels to be used
    |                           for the calculation. Must be integers >= 1.
    |                           Default is all channels 1 to 10.
    |   center_atoms (AtomSelection): AtomSelection, integer index or list of
    |                                 indices of all atoms around which the
    |                                 bond order parameters are to be computed
    |                                 and summed over. Default is None, which
    |                                 means all atoms of the system.
    |   environment_atoms (AtomSelection): AtomSelection, integer index or
    |                                      list of indices of all atoms that
    |                                      used as environment to compute the
    |                                      bond order parameters. For each
    |                                      'center' atom, all its links to
    |                                      'environment' atoms other than
    |                                      itself will be considered. Default
    |                                      is None, which means all atoms of
    |                                      the system.
    |   cutoff_radius (float): radius at which to cut the smoothing sigmoid
    |                          function, in Angstroms. Default is 2.0.
    |   cutoff_width (float): width parameter of the sigmoid, controlling the
    |                         steepness with which it falls to zero, in 
    |                         Angstroms. Default is 0.05.
    |   compute_W (bool): whether to also compute the W parameter (third
    |                     order) besides the Q (second order).

    | Returns:
    |   bond_order_pars (dict): dictionary containing arrays with Q and, if
    |                           required, W for each l in l_channels.

    """

    default_name = 'bond_order'
    default_params = {
        'l_channels': np.arange(1, 11),
        'center_atoms': None,
        'environment_atoms': None,
        'cutoff_radius': 2.0,
        'cutoff_width': 0.05,
        'compute_W': False
    }

    @staticmethod
    def extract(s, l_channels, center_atoms, environment_atoms, cutoff_radius,
                cutoff_width, compute_W):

        # Check that l_channels are valid
        l_channels = np.array(l_channels)
        if ((l_channels < 1) | ((l_channels % 1) != 0)).any():
            raise ValueError('Invalid angular momentum channels')

        # Turn center_atoms and environment_atoms into AtomSelections
        if not isinstance(center_atoms, AtomSelection):
            if center_atoms is None:
                center_atoms = range(len(s))
            elif not isinstance(center_atoms, list):
                center_atoms = [center_atoms]
            center_atoms = AtomSelection(s, center_atoms)

        if not isinstance(environment_atoms, AtomSelection):
            if environment_atoms is None:
                environment_atoms = range(len(s))
            elif not isinstance(environment_atoms, list):
                environment_atoms = [environment_atoms]
            environment_atoms = AtomSelection(s, environment_atoms)

        xyz = s.get_positions()

        # Now to build the total list of vectors and weights
        all_vecs = np.zeros((0, 3))
        all_weights = np.zeros((0,))

        for ci in center_atoms.indices:
            # What would the environment be?
            env_inds = list(set(environment_atoms.indices) - set([ci]))
            # So what are the 'bonds'?
            bonds = minimum_periodic(xyz[env_inds] - xyz[ci], s.get_cell())[0]
            bnorms = np.linalg.norm(bonds, axis=1)

            # Compute sigmoid weights
            bweights = 0.5*(((cutoff_radius-bnorms)/cutoff_width) /
                            (((cutoff_radius-bnorms)/cutoff_width)**2+1)**0.5
                            + 1)
            bweights /= np.sum(bweights)  # Norm to 1

            all_vecs = np.concatenate((all_vecs, bonds))
            all_weights = np.concatenate((all_weights, bweights))

        # Norm to 1 all weights:
        all_weights /= len(center_atoms.indices)

        # Now compute the order parameters!
        stp = _steinhardt_pars(all_vecs, l_channels, weights=all_weights,
                               compute_W=compute_W)

        if compute_W:
            return {'Q': stp[0], 'W': stp[1]}
        else:
            return {'Q': stp}



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Generator producing structures interpolated between two extremes"""

import os
import numpy as np
# Internal imports
import soprano.utils as utils


def linspaceGen(struct_0, struct_1, steps=10, periodic=False):
    """Generator function to create multiple structures with positions
    interpolated linearly between two extremes.

    | Args:
    |   struct_0 (ase.Atoms): the starting structure
    |   struct_1 (ase.Atoms): the final structure. The atoms should be in the
    |                         same order as the ones in struct_0
    |   steps (Optional[int]): number of interpolated steps to produce
    |                          (extremes included). Default is 10
    |   periodic (Optional[bool]): if True the interpolation will take into
    |                              account periodic boundaries and interpolate
    |                              between positions in struct_0 and the
    |                              closest periodic copy of positions in
    |                              struct_1. By default set to False

    | Returns:
    |   linspaceGenerator (generator): an iterator object that yields
    |                                  structures created by linear
    |                                  interpolation.

    """

    # First, a compatibility check
    chem0 = struct_0.get_chemical_symbols()
    chem1 = struct_1.get_chemical_symbols()

    if chem0 != chem1:
        raise RuntimeError('The two structures passed to linspaceGen do not '
                           'have the same chemical composition')

    pos0 = struct_0.get_positions()
    pos1 = struct_1.get_positions()

    rootname = struct_0.info['name'] if 'name' in struct_0.info else 'linspace'

    # Adjust pos1 to be periodic if asked to
    if periodic:
        dpos = pos1-pos0
        dpos = utils.minimum_periodic(dpos, struct_0.get_cell())[0]
        pos1 = pos0+dpos

    for i, t in enumerate(np.linspace(0, 1, steps)):

        pos = pos0*(1-t)+pos1*t
        struct = struct_0.copy()
        struct.set_positions(pos)
        struct.info['name'] = '{0}_{1}'.format(rootname, i)

        yield struct
